from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np
import torch
from matplotlib.patches import Rectangle
from scipy import constants
from scipy.stats import multivariate_normal

from cheetah import utils
from cheetah.particles import Beam, ParameterBeam, ParticleBeam

ELEMENT_COUNT = 0
REST_ENERGY = (
    constants.electron_mass
    * constants.speed_of_light**2
    / constants.elementary_charge
)


class Element:
    """
    Base class for elements of particle accelerators.

    Parameters
    ----------
    name : string, optional
        Unique identifier of the element.

    Attributes
    ---------
    is_active : bool
        Is set to `True` when the element is in operation. May be defined differently
        for each type of element.
    is_skippable : bool
        Marking an element as skippable allows the transfer map to be combined with
        those of preceeding and succeeding elements and results in faster particle
        tracking. This property has to be defined by subclasses of `Element` and made be
        set dynamically depending on their current mode of operation.
    device : string
        Device to move the beam's particle array to. If set to `"auto"` a CUDA GPU is
        selected if available. The CPU is used otherwise.
    """

    is_active = False
    is_skippable = True

    def __init__(self, name=None, device="auto"):
        global ELEMENT_COUNT
        if name is not None:
            self.name = name
        else:
            self.name = f"{self.__class__.__name__}_{ELEMENT_COUNT:06d}"
        ELEMENT_COUNT += 1

        if device == "auto":
            device = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = device

    def transfer_map(self, energy):
        raise NotImplementedError

    def __call__(self, incoming):
        """
        Track particles through the element.

        Pramameters
        -----------
        incoming : cheetah.Beam
            Beam of particles entering the element.

        Returns
        -------
        cheetah.Beam
            Beam of particles exiting the element.
        """
        if incoming is Beam.empty:
            return incoming
        elif isinstance(incoming, ParameterBeam):
            tm = self.transfer_map(incoming.energy)
            mu = torch.matmul(tm, incoming._mu)
            cov = torch.matmul(tm, torch.matmul(incoming._cov, tm.t()))
            return ParameterBeam(mu, cov, incoming.energy)
        elif isinstance(incoming, ParticleBeam):
            tm = self.transfer_map(incoming.energy)
            new_particles = torch.matmul(incoming.particles, tm.t())
            return ParticleBeam(new_particles, incoming.energy, device=incoming.device)
        else:
            raise TypeError(f"Parameter incoming is of invalid type {type(incoming)}")

    def split(self, resolution):
        """
        Split the element into slices no longer than `resolution`.

        Parameters
        ----------
        resolution : float
            Length of the longest allowed split in meters.

        Returns
        -------
        list
            Ordered sequence of sliced elements.

        Raises
        ------
        NotImplementedError
            If not split function is implemented for the given `Element` subclass.
        """
        raise NotImplementedError

    def plot(self, ax, s):
        """
        Plot a representation of this element into a `matplotlib` Axes at position `s`.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes to plot the representation into.
        s : float
            Position of the object along s in meters.

        Raises
        ------
        NotImplementedError
            If not split function is implemented for the given `Element` subclass.
        """
        raise NotImplementedError

    def __repr__(self):
        return f'{self.__class__.__name__}(name="{self.name}")'


class Drift(Element):
    """
    Drift section in a particle accelerator.

    Parameters
    ----------
    length : float
        Length in meters.
    energy : float, optional
    name : string, optional
        Unique identifier of the element.

    Attributes
    ---------
    is_active : bool
        Drifts are always set as active.
    """

    is_active = True
    is_skippable = True

    def __init__(self, length, name=None, **kwargs):
        self.length = length

        super().__init__(name=name, **kwargs)

    def transfer_map(self, energy):
        gamma = energy / REST_ENERGY
        igamma2 = 1 / gamma**2 if gamma != 0 else 0

        return torch.tensor(
            [
                [1, self.length, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, 0],
                [0, 0, 1, self.length, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, 0],
                [0, 0, 0, 0, 1, self.length * igamma2, 0],
                [0, 0, 0, 0, 0, 1, 0],
                [0, 0, 0, 0, 0, 0, 1],
            ],
            dtype=torch.float32,
            device=self.device,
        )

    def split(self, resolution):
        split_elements = []
        remaining = self.length
        while remaining > 0:
            element = Drift(min(resolution, remaining), device=self.device)
            split_elements.append(element)
            remaining -= resolution
        return split_elements

    def plot(self, ax, s):
        pass

    def __repr__(self):
        return (
            f'{self.__class__.__name__}(length={self.length:.2f}, name="{self.name}")'
        )


class Quadrupole(Element):
    """
    Quadrupole magnet in a particle accelerator.

    Parameters
    ----------
    length : float
        Length in meters.
    k1 : float, optional
        Strength of the quadrupole in rad/m.
    misalignment : (float, float), optional
        Misalignment vector of the quadrupole in x- and y-directions.
    energy : float, optional
    name : string, optional
        Unique identifier of the element.

    Attributes
    ---------
    is_active : bool
        Is set `True` when `k1 != 0`.
    """

    is_skippable = True

    def __init__(self, length, k1=0.0, misalignment=(0, 0), name=None, **kwargs):
        self.length = length
        self.k1 = k1
        self.misalignment = misalignment

        super().__init__(name=name, **kwargs)

    def transfer_map(self, energy):
        gamma = energy / REST_ENERGY
        igamma2 = 1 / gamma**2 if gamma != 0 else 0

        beta = np.sqrt(1 - igamma2)

        hx = 0
        kx2 = self.k1 + hx**2
        ky2 = -self.k1
        kx = np.sqrt(kx2 + 0.0j)
        ky = np.sqrt(ky2 + 0.0j)
        cx = np.cos(kx * self.length).real
        cy = np.cos(ky * self.length).real
        sy = (np.sin(ky * self.length) / ky).real if ky != 0 else self.length

        if kx != 0:
            sx = (np.sin(kx * self.length) / kx).real
            dx = hx / kx2 * (1.0 - cx)
            r56 = hx**2 * (self.length - sx) / kx2 / beta**2
        else:
            sx = self.length
            dx = self.length**2 * hx / 2
            r56 = hx**2 * self.length**3 / 6 / beta**2

        r56 -= self.length / beta**2 * igamma2

        R = torch.tensor(
            [
                [cx, sx, 0, 0, 0, dx / beta, 0],
                [-kx2 * sx, cx, 0, 0, 0, sx * hx / beta, 0],
                [0, 0, cy, sy, 0, 0, 0],
                [0, 0, -ky2 * sy, cy, 0, 0, 0],
                [sx * hx / beta, dx / beta, 0, 0, 1, r56, 0],
                [0, 0, 0, 0, 0, 1, 0],
                [0, 0, 0, 0, 0, 0, 1],
            ],
            dtype=torch.float32,
            device=self.device,
        )

        if self.misalignment[0] == 0 and self.misalignment[1] == 0:
            return R
        else:
            R_exit = torch.tensor(
                [
                    [1, 0, 0, 0, 0, 0, self.misalignment[0]],
                    [0, 1, 0, 0, 0, 0, 0],
                    [0, 0, 1, 0, 0, 0, self.misalignment[1]],
                    [0, 0, 0, 1, 0, 0, 0],
                    [0, 0, 0, 0, 1, 0, 0],
                    [0, 0, 0, 0, 0, 1, 0],
                    [0, 0, 0, 0, 0, 0, 1],
                ],
                dtype=torch.float32,
                device=self.device,
            )
            R_entry = torch.tensor(
                [
                    [1, 0, 0, 0, 0, 0, -self.misalignment[0]],
                    [0, 1, 0, 0, 0, 0, 0],
                    [0, 0, 1, 0, 0, 0, -self.misalignment[1]],
                    [0, 0, 0, 1, 0, 0, 0],
                    [0, 0, 0, 0, 1, 0, 0],
                    [0, 0, 0, 0, 0, 1, 0],
                    [0, 0, 0, 0, 0, 0, 1],
                ],
                dtype=torch.float32,
                device=self.device,
            )
            R = torch.matmul(R_exit, torch.matmul(R, R_entry))
            return R

    @property
    def is_active(self):
        return self.k1 != 0

    def split(self, resolution):
        split_elements = []
        remaining = self.length
        while remaining > 0:
            element = Quadrupole(
                min(resolution, remaining),
                self.k1,
                misalignment=self.misalignment,
                device=self.device,
            )
            split_elements.append(element)
            remaining -= resolution
        return split_elements

    def plot(self, ax, s):
        alpha = 1 if self.is_active else 0.2
        height = 0.8 * (np.sign(self.k1) if self.is_active else 1)
        patch = Rectangle(
            (s, 0), self.length, height, color="tab:red", alpha=alpha, zorder=2
        )
        ax.add_patch(patch)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(length={self.length:.2f}, "
            + f"k1={self.k1}, "
            + f"misalignment={self.misalignment}, "
            + f'name="{self.name}")'
        )


class HorizontalCorrector(Element):
    """
    Horizontal corrector magnet in a particle accelerator.

    Parameters
    ----------
    length : float
        Length in meters.
    angle : float, optional
        Particle deflection angle in the horizontal plane in rad.
    energy : float, optional
    name : string, optional
        Unique identifier of the element.

    Attributes
    ---------
    is_active : bool
        Is set `True` when `angle != 0`.
    """

    is_skippable = True

    def __init__(self, length, angle=0.0, name=None, **kwargs):
        self.length = length
        self.angle = angle

        super().__init__(name=name, **kwargs)

    def transfer_map(self, energy):
        return torch.tensor(
            [
                [1, self.length, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, self.angle],
                [0, 0, 1, self.length, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, 0],
                [0, 0, 0, 0, 1, 0, 0],
                [0, 0, 0, 0, 0, 1, 0],
                [0, 0, 0, 0, 0, 0, 1],
            ],
            dtype=torch.float32,
            device=self.device,
        )

    @property
    def is_active(self):
        return self.angle != 0

    def split(self, resolution):
        split_elements = []
        remaining = self.length
        while remaining > 0:
            length = min(resolution, remaining)
            element = HorizontalCorrector(
                length, self.angle * length / self.length, device=self.device
            )
            split_elements.append(element)
            remaining -= resolution
        return split_elements

    def plot(self, ax, s):
        alpha = 1 if self.is_active else 0.2
        height = 0.8 * (np.sign(self.angle) if self.is_active else 1)

        patch = Rectangle(
            (s, 0), self.length, height, color="tab:blue", alpha=alpha, zorder=2
        )
        ax.add_patch(patch)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(length={self.length:.2f}, "
            + f"angle={self.angle}, "
            + f'name="{self.name}")'
        )


class VerticalCorrector(Element):
    """
    Verticle corrector magnet in a particle accelerator.

    Parameters
    ----------
    length : float
        Length in meters.
    angle : float, optional
        Particle deflection angle in the vertical plane in rad.
    energy : float, optional
    name : string, optional
        Unique identifier of the element.

    Attributes
    ---------
    is_active : bool
        Is set `True` when `angle != 0`.
    """

    is_skippable = True

    def __init__(self, length, angle=0.0, name=None, **kwargs):
        self.length = length
        self.angle = angle

        super().__init__(name=name, **kwargs)

    def transfer_map(self, energy):
        return torch.tensor(
            [
                [1, self.length, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, 0],
                [0, 0, 1, self.length, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, self.angle],
                [0, 0, 0, 0, 1, 0, 0],
                [0, 0, 0, 0, 0, 1, 0],
                [0, 0, 0, 0, 0, 0, 1],
            ],
            dtype=torch.float32,
            device=self.device,
        )

    @property
    def is_active(self):
        return self.angle != 0

    def split(self, resolution):
        split_elements = []
        remaining = self.length
        while remaining > 0:
            length = min(resolution, remaining)
            element = VerticalCorrector(
                length, self.angle * length / self.length, device=self.device
            )
            split_elements.append(element)
            remaining -= resolution
        return split_elements

    def plot(self, ax, s):
        alpha = 1 if self.is_active else 0.2
        height = 0.8 * (np.sign(self.angle) if self.is_active else 1)

        patch = Rectangle(
            (s, 0), self.length, height, color="tab:cyan", alpha=alpha, zorder=2
        )
        ax.add_patch(patch)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(length={self.length:.2f}, "
            + f"angle={self.angle}, "
            + f'name="{self.name}")'
        )


class Cavity(Element):
    """
    Accelerating cavity in a particle accelerator.

    Parameters
    ----------
    length : float
        Length in meters.
    delta_energy : float, optional
        Energy added to the beam by the accelerating cavity.
    name : string, optional
        Unique identifier of the element.
    """

    def __init__(self, length, delta_energy=0, name=None, **kwargs):
        self.length = length
        self.delta_energy = delta_energy

        super().__init__(name=name, **kwargs)

    @property
    def is_active(self):
        return self.delta_energy != 0

    @property
    def is_skippable(self):
        return not self.is_active

    def transfer_map(self, energy):
        gamma = energy / REST_ENERGY
        igamma2 = 1 / gamma**2 if gamma != 0 else 0

        return torch.tensor(
            [
                [1, self.length, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, 0],
                [0, 0, 1, self.length, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, 0],
                [0, 0, 0, 0, 1, self.length * igamma2, 0],
                [0, 0, 0, 0, 0, 1, 0],
                [0, 0, 0, 0, 0, 0, 1],
            ],
            dtype=torch.float32,
            device=self.device,
        )

    def __call__(self, incoming):
        outgoing = super().__call__(incoming)
        if outgoing is not Beam.empty:
            outgoing.energy += self.delta_energy
        return outgoing

    def split(self, resolution):
        split_elements = []
        remaining = self.length
        while remaining > 0:
            split_length = min(resolution, remaining)
            split_delta_energy = self.delta_energy * split_length / self.length
            element = Cavity(
                split_length, delta_energy=split_delta_energy, device=self.device
            )
            split_elements.append(element)
            remaining -= resolution
        return split_elements

    def plot(self, ax, s):
        alpha = 1 if self.is_active else 0.2
        height = 0.4

        patch = Rectangle(
            (s, 0), self.length, height, color="gold", alpha=alpha, zorder=2
        )
        ax.add_patch(patch)

    def __repr__(self):
        return f'{self.__class__.__name__}(length={self.length:.2f}, delta_energy={self.delta_energy}, name="{self.name}")'


class BPM(Element):
    """
    Beam Position Monitor (BPM) in a particle accelerator.

    Parameters
    ----------
    name : string, optional
        Unique identifier of the element.

    Attributes
    ---------
    is_active : bool
        Can be set by the user. Merely influences how the element is displayed in a
        lattice plot.
    reading : (float, float)
        Beam position read by the BPM. Is refreshed when the BPM is active and a beam is
        tracked through it. Before tracking a beam through here, the reading is
        initialised as `(None, None)`.
    """

    length = 0
    is_skippable = True  # TODO: Temporary

    reading = (None, None)

    @property
    def is_skippable(self):
        return not self.is_active

    def transfer_map(self, energy):
        return torch.eye(7, device=self.device)

    def __call__(self, incoming):
        if incoming is Beam.empty:
            self.reading = (None, None)
            return Beam.empty
        else:
            self.reading = (incoming.mu_x, incoming.mu_y)
            return ParticleBeam(incoming.particles, incoming.energy, device=self.device)

    def split(self, resolution):
        return [self]

    def plot(self, ax, s):
        alpha = 1 if self.is_active else 0.2
        patch = Rectangle(
            (s, -0.3), 0, 0.3 * 2, color="darkkhaki", alpha=alpha, zorder=2
        )
        ax.add_patch(patch)


class Screen(Element):
    """
    Diagnostic screen in a particle accelerator.

    Parameters
    ----------
    name : string, optional
        Unique identifier of the element.
    resolution : (int, int)
        Resolution of the camera sensor looking at the screen given as a tuple
        `(width, height)`.
    binning : int, optional
        Binning used by the camera.

    Attributes
    ---------
    is_active : bool
        Can be set by the user. An active screen records an image and blocks all
        particles when a beam is tracked through it.
    """

    length = 0

    def __init__(
        self,
        resolution,
        pixel_size,
        binning=1,
        misalignment=(0, 0),
        name=None,
        **kwargs,
    ):
        super().__init__(name=name, **kwargs)

        self.resolution = resolution
        self.pixel_size = pixel_size
        self.binning = binning
        self.misalignment = misalignment

        self.read_beam = None
        self.cached_reading = None

    @property
    def is_skippable(self):
        return not self.is_active

    @property
    def effective_resolution(self):
        return (
            int(self.resolution[0] / self.binning),
            int(self.resolution[1] / self.binning),
        )

    @property
    def effective_pixel_size(self):
        return (self.pixel_size[0] * self.binning, self.pixel_size[0] * self.binning)

    @property
    def extent(self):
        return (
            -self.resolution[0] * self.pixel_size[0] / 2,
            self.resolution[0] * self.pixel_size[0] / 2,
            -self.resolution[1] * self.pixel_size[1] / 2,
            self.resolution[1] * self.pixel_size[1] / 2,
        )

    @property
    def pixel_bin_edges(self):
        return (
            torch.linspace(
                -self.resolution[0] * self.pixel_size[0] / 2,
                self.resolution[0] * self.pixel_size[0] / 2,
                self.effective_resolution[0] + 1,
            ),
            torch.linspace(
                -self.resolution[1] * self.pixel_size[1] / 2,
                self.resolution[1] * self.pixel_size[1] / 2,
                self.effective_resolution[1] + 1,
            ),
        )

    def transfer_map(self, energy):
        return torch.eye(7, device=self.device)

    def __call__(self, incoming):
        if self.is_active:
            if isinstance(incoming, ParameterBeam):
                self.read_beam = deepcopy(incoming)
                self.read_beam._mu[0] -= self.misalignment[0]
                self.read_beam._mu[2] -= self.misalignment[1]
            elif isinstance(incoming, ParticleBeam):
                self.read_beam = deepcopy(incoming)
                x_offset = np.full(len(self.read_beam), self.misalignment[0])
                y_offset = np.full(len(self.read_beam), self.misalignment[1])
                self.read_beam.particles[:, 0] -= x_offset
                self.read_beam.particles[:, 1] -= y_offset
            else:
                self.read_beam = incoming

            return Beam.empty
        else:
            return incoming

    @property
    def reading(self):
        if self.cached_reading is not None:
            return self.cached_reading

        if self.read_beam is Beam.empty or self.read_beam is None:
            image = torch.zeros(
                (self.effective_resolution[1], self.effective_resolution[0])
            )
        elif isinstance(self.read_beam, ParameterBeam):
            transverse_mu = np.array([self.read_beam._mu[0], self.read_beam._mu[2]])
            transverse_cov = np.array(
                [
                    [self.read_beam._cov[0, 0], self.read_beam._cov[0, 2]],
                    [self.read_beam._cov[2, 0], self.read_beam._cov[2, 2]],
                ]
            )
            dist = multivariate_normal(mean=transverse_mu, cov=transverse_cov)

            left = self.extent[0]
            right = self.extent[1]
            hstep = self.pixel_size[0] * self.binning
            bottom = self.extent[2]
            top = self.extent[3]
            vstep = self.pixel_size[1] * self.binning
            x, y = np.mgrid[left:right:hstep, bottom:top:vstep]
            pos = np.dstack((x, y))
            image = dist.pdf(pos)
            image = np.flipud(image.T)
        elif isinstance(self.read_beam, ParticleBeam):
            image, _ = torch.histogramdd(
                torch.stack((self.read_beam.xs, self.read_beam.ys)),
                bins=self.pixel_bin_edges,
            )
            image = torch.flipud(image.T)
            image = image.cpu()
        else:
            raise TypeError(f"Read beam is of invalid type {type(self.read_beam)}")

        self.cached_reading = image
        return image

    @property
    def read_beam(self):
        return self._read_beam

    @read_beam.setter
    def read_beam(self, value):
        self._read_beam = value
        self.cached_reading = None

    def split(self, resolution):
        return [self]

    def plot(self, ax, s):
        alpha = 1 if self.is_active else 0.2
        patch = Rectangle(
            (s, -0.6), 0, 0.6 * 2, color="tab:green", alpha=alpha, zorder=2
        )
        ax.add_patch(patch)

    def __repr__(self):
        return f'{self.__class__.__name__}(resolution={self.resolution}, pixel_size={self.pixel_size}, binning={self.binning}, misalignment={self.misalignment}, name="{self.name}")'


class Undulator(Element):
    """
    Element representing an undulator in a particle accelerator.

    Parameters
    ----------
    length : float
        Length in meters.
    name : string, optional
        Unique identifier of the element.

    Notes
    -----
    Currently behaves like a drift section but is plotted distinctively.
    """

    is_skippable = True  # TODO: Temporary?

    def __init__(self, length, name=None, **kwargs):
        self.length = length

        super().__init__(name=name, **kwargs)

    def transfer_map(self, energy):
        gamma = energy / REST_ENERGY
        igamma2 = 1 / gamma**2 if gamma != 0 else 0

        return torch.tensor(
            [
                [1, self.length, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, 0],
                [0, 0, 1, self.length, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, 0],
                [0, 0, 0, 0, 1, self.length * igamma2, 0],
                [0, 0, 0, 0, 0, 1, 0],
                [0, 0, 0, 0, 0, 0, 1],
            ],
            dtype=torch.float32,
            device=self.device,
        )

    def split(self, resolution):
        split_elements = []
        remaining = self.length
        while remaining > 0:
            element = Cavity(min(resolution, remaining), device=self.device)
            split_elements.append(element)
            remaining -= resolution
        return split_elements

    def plot(self, ax, s):
        alpha = 1 if self.is_active else 0.2
        height = 0.4

        patch = Rectangle(
            (s, 0), self.length, height, color="tab:purple", alpha=alpha, zorder=2
        )
        ax.add_patch(patch)

    def __repr__(self):
        return (
            f'{self.__class__.__name__}(length={self.length:.2f}, name="{self.name}")'
        )


class Segment(Element):
    """
    Segment of a particle accelerator consisting of several elements.

    Parameters
    ----------
    cell : list
        List of Cheetah elements that describe an accelerator (section).
    name : string, optional
        Unique identifier of the element.
    """

    def __init__(self, cell, name=None, **kwargs):
        super().__init__(name=name, **kwargs)

        self.elements = cell

        for element in self.elements:
            element.device = self.device
            self.__dict__[element.name] = element

    def subcell(self, start, end, **kwargs):
        """Extract a subcell `[start, end]` from an this segment."""
        subcell = []
        is_in_subcell = False
        for element in self.elements:
            if element.name == start:
                is_in_subcell = True
            if is_in_subcell:
                subcell.append(element)
            if element.name == end:
                break

        return self.__class__(subcell, device=self.device, **kwargs)

    @classmethod
    def from_ocelot(cls, cell, name=None, warnings=True, **kwargs):
        converted = [
            utils.ocelot2cheetah(element, warnings=warnings) for element in cell
        ]
        return cls(converted, name=name, **kwargs)

    @property
    def is_skippable(self):
        return all(element.is_skippable for element in self.elements)

    @property
    def length(self):
        return sum(element.length for element in self.elements)

    def transfer_map(self, energy):
        if self.is_skippable:
            tm = torch.eye(7, dtype=torch.float32, device=self.device)
            for element in self.elements:
                tm = torch.matmul(element.transfer_map(energy), tm)
            return tm
        else:
            return None

    def __call__(self, incoming):
        if self.is_skippable:
            return super().__call__(incoming)
        else:
            todos = []
            for element in self.elements:
                if not element.is_skippable:
                    todos.append(element)
                elif not todos or not todos[-1].is_skippable:
                    todos.append(Segment([element], device=self.device))
                else:
                    todos[-1].elements.append(element)

            for todo in todos:
                incoming = todo(incoming)

            return incoming

    def split(self, resolution):
        return [
            split_element
            for element in self.elements
            for split_element in element.split(resolution)
        ]

    def plot(self, ax, s):
        element_lengths = [element.length for element in self.elements]
        element_ss = [0] + [
            sum(element_lengths[: i + 1]) for i, _ in enumerate(element_lengths)
        ]
        element_ss = [s + element_s for element_s in element_ss]

        ax.plot([0, element_ss[-1]], [0, 0], "--", color="black")

        for element, s in zip(self.elements, element_ss[:-1]):
            element.plot(ax, s)

        ax.set_ylim(-1, 1)
        ax.set_xlabel("s (m)")
        ax.set_yticks([])
        ax.grid()

    def plot_reference_particle_traces(
        self, axx, axy, beam=None, n=10, resolution=0.01
    ):
        """
        Plot `n` reference particles along the segment view in x- and y-direction.

        Parameters
        ----------
        axx : matplotlib.axes.Axes
            Axes to plot the particle traces into viewed in x-direction.
        axy : matplotlib.axes.Axes
            Axes to plot the particle traces into viewed in y-direction.
        beam : cheetah.Beam, optional
            Entering beam from which the reference particles are sampled.
        n : int, optional
            Number of reference particles to plot. Must not be larger than number of
            particles passed in `particles`.
        resolution : float, optional
            Minimum resolution of the tracking of the reference particles in the plot.
        """
        reference_segment = deepcopy(self)
        splits = reference_segment.split(resolution)

        split_lengths = [split.length for split in splits]
        ss = [0] + [sum(split_lengths[: i + 1]) for i, _ in enumerate(split_lengths)]

        references = []
        if beam is None:
            initial = ParticleBeam.make_linspaced(n=n, device="cpu")
            references.append(initial)
        else:
            initial = ParticleBeam.make_linspaced(
                n=n,
                mu_x=beam.mu_x,
                mu_xp=beam.mu_xp,
                mu_y=beam.mu_y,
                mu_yp=beam.mu_yp,
                sigma_x=beam.sigma_x,
                sigma_xp=beam.sigma_xp,
                sigma_y=beam.sigma_y,
                sigma_yp=beam.sigma_yp,
                sigma_s=beam.sigma_s,
                sigma_p=beam.sigma_p,
                energy=beam.energy,
                device="cpu",
            )
            references.append(initial)
        for split in splits:
            sample = split(references[-1])
            references.append(sample)

        for particle_index in range(n):
            xs = [
                float(reference_beam.xs[particle_index].cpu())
                for reference_beam in references
                if reference_beam is not Beam.empty
            ]
            axx.plot(ss[: len(xs)], xs)
        axx.set_xlabel("s (m)")
        axx.set_ylabel("x (m)")
        axx.grid()

        for particle_index in range(n):
            ys = [
                float(reference_beam.ys[particle_index].cpu())
                for reference_beam in references
                if reference_beam is not Beam.empty
            ]
            axy.plot(ss[: len(ys)], ys)
        axx.set_xlabel("s (m)")
        axy.set_ylabel("y (m)")
        axy.grid()

    def plot_overview(self, fig=None, beam=None, n=10, resolution=0.01):
        """
        Plot an overview of the segment with the lattice and traced reference particles.

        Parameters
        ----------
        fig: matplotlib.figure.Figure, optional
            Figure to plot the overview into.
        beam : cheetah.Beam, optional
            Entering beam from which the reference particles are sampled.
        n : int, optional
            Number of reference particles to plot. Must not be larger than number of
            particles passed in `beam`.
        resolution : float, optional
            Minimum resolution of the tracking of the reference particles in the plot.
        """
        if fig is None:
            fig = plt.figure()
        gs = fig.add_gridspec(3, hspace=0, height_ratios=[2, 2, 1])
        axs = gs.subplots(sharex=True)

        axs[0].set_title("Reference Particle Traces")
        self.plot_reference_particle_traces(axs[0], axs[1], beam, n, resolution)

        self.plot(axs[2], 0)

        plt.tight_layout()

    def __repr__(self):
        start = f"{self.__class__.__name__}(["

        s = start + self.elements[0].__repr__()
        x = [", " + element.__repr__() for element in self.elements[1:]]
        s += "".join(x)
        s += "])"

        return s



import numpy as np
import torch
from torch.distributions import MultivariateNormal

from cheetah.utils import from_astrabeam


class Beam:

    empty = "I'm an empty beam!"

    @classmethod
    def from_parameters(
        cls,
        mu_x=0,
        mu_xp=0,
        mu_y=0,
        mu_yp=0,
        sigma_x=175e-9,
        sigma_xp=2e-7,
        sigma_y=175e-9,
        sigma_yp=2e-7,
        sigma_s=1e-6,
        sigma_p=1e-6,
        cor_x=0,
        cor_y=0,
        cor_s=0,
        energy=1e8,
    ):
        """
        Create beam that with given beam parameters.

        Parameters
        ----------
        n : int, optional
            Number of particles to generate.
        mu_x : float, optional
            Center of the particle distribution on x in meters.
        mu_xp : float, optional
            Center of the particle distribution on px in meters.
        mu_y : float, optional
            Center of the particle distribution on y in meters.
        mu_yp : float, optional
            Center of the particle distribution on py in meters.
        sigma_x : float, optional
            Sgima of the particle distribution in x direction in meters.
        sigma_xp : float, optional
            Sgima of the particle distribution in px direction in meters.
        sigma_y : float, optional
            Sgima of the particle distribution in y direction in meters.
        sigma_yp : float, optional
            Sgima of the particle distribution in py direction in meters.
        sigma_s : float, optional
            Sgima of the particle distribution in s direction in meters.
        sigma_p : float, optional
            Sgima of the particle distribution in p direction in meters.
        energy : float, optional
            Energy of the beam in eV.
        """
        raise NotImplementedError

    @classmethod
    def from_ocelot(cls, parray):
        """
        Convert an Ocelot ParticleArray `parray` to a Cheetah Beam.
        """
        raise NotImplementedError

    @classmethod
    def from_astra(cls, path, **kwargs):
        """Load an Astra particle distribution as a Cheetah Beam."""
        raise NotImplementedError

    def transformed_to(
        self,
        mu_x=None,
        mu_xp=None,
        mu_y=None,
        mu_yp=None,
        sigma_x=None,
        sigma_xp=None,
        sigma_y=None,
        sigma_yp=None,
        sigma_s=None,
        sigma_p=None,
        energy=None,
    ):
        """
        Create version of this beam that is transformed to new beam parameters.

        Parameters
        ----------
        n : int, optional
            Number of particles to generate.
        mu_x : float, optional
            Center of the particle distribution on x in meters.
        mu_xp : float, optional
            Center of the particle distribution on px in meters.
        mu_y : float, optional
            Center of the particle distribution on y in meters.
        mu_yp : float, optional
            Center of the particle distribution on py in meters.
        sigma_x : float, optional
            Sgima of the particle distribution in x direction in meters.
        sigma_xp : float, optional
            Sgima of the particle distribution in px direction in meters.
        sigma_y : float, optional
            Sgima of the particle distribution in y direction in meters.
        sigma_yp : float, optional
            Sgima of the particle distribution in py direction in meters.
        sigma_s : float, optional
            Sgima of the particle distribution in s direction in meters.
        sigma_p : float, optional
            Sgima of the particle distribution in p direction in meters.
        energy : float, optional
            Energy of the beam in eV.
        """
        mu_x = mu_x if mu_x != None else self.mu_x
        mu_xp = mu_xp if mu_xp != None else self.mu_xp
        mu_y = mu_y if mu_y != None else self.mu_y
        mu_yp = mu_yp if mu_yp != None else self.mu_yp
        sigma_x = sigma_x if sigma_x != None else self.sigma_x
        sigma_xp = sigma_xp if sigma_xp != None else self.sigma_xp
        sigma_y = sigma_y if sigma_y != None else self.sigma_y
        sigma_yp = sigma_yp if sigma_yp != None else self.sigma_yp
        sigma_s = sigma_s if sigma_s != None else self.sigma_s
        sigma_p = sigma_p if sigma_p != None else self.sigma_p
        energy = energy if energy != None else self.energy

        return self.__class__.from_parameters(
            mu_x=mu_x,
            mu_xp=mu_xp,
            mu_y=mu_y,
            mu_yp=mu_yp,
            sigma_x=sigma_x,
            sigma_xp=sigma_xp,
            sigma_y=sigma_y,
            sigma_yp=sigma_yp,
            sigma_s=sigma_s,
            sigma_p=sigma_p,
            energy=energy,
        )

    @property
    def parameters(self):
        return {
            "mu_x": self.mu_x,
            "mu_xp": self.mu_xp,
            "mu_y": self.mu_y,
            "mu_yp": self.mu_yp,
            "sigma_x": self.sigma_x,
            "sigma_xp": self.sigma_xp,
            "sigma_y": self.sigma_y,
            "sigma_yp": self.sigma_yp,
            "sigma_s": self.sigma_s,
            "sigma_p": self.sigma_p,
            "energy": self.energy,
        }

    @property
    def mu_x(self):
        raise NotImplementedError

    @property
    def sigma_x(self):
        raise NotImplementedError

    @property
    def mu_xp(self):
        raise NotImplementedError

    @property
    def sigma_xp(self):
        raise NotImplementedError

    @property
    def mu_y(self):
        raise NotImplementedError

    @property
    def sigma_y(self):
        raise NotImplementedError

    @property
    def mu_yp(self):
        raise NotImplementedError

    @property
    def sigma_yp(self):
        raise NotImplementedError

    @property
    def mu_s(self):
        raise NotImplementedError

    @property
    def sigma_s(self):
        raise NotImplementedError

    @property
    def mu_p(self):
        raise NotImplementedError

    @property
    def sigma_p(self):
        raise NotImplementedError

    def __repr__(self):
        return f"{self.__class__.__name__}(mu_x={self.mu_x}, mu_xp={self.mu_xp}, mu_y={self.mu_y}, mu_yp={self.mu_yp}, sigma_x={self.sigma_x}, sigma_xp={self.sigma_xp}, sigma_y={self.sigma_y}, sigma_yp={self.sigma_yp}, sigma_s={self.sigma_s}, sigma_p={self.sigma_p}, energy={self.energy})"


class ParameterBeam(Beam):
    """
    Beam of charged particles, where each particle is simulated.

    Parameters
    ----------
    mu : torch.Tensor
        Mu vector of the beam.
    cov : torch.Tensor
        Covariance matrix of the beam.
    energy : float
        Energy of the beam in eV.
    """

    def __init__(self, mu, cov, energy):
        self._mu = mu
        self._cov = cov
        self.energy = energy

    @classmethod
    def from_parameters(
        cls,
        mu_x=0,
        mu_xp=0,
        mu_y=0,
        mu_yp=0,
        sigma_x=175e-9,
        sigma_xp=2e-7,
        sigma_y=175e-9,
        sigma_yp=2e-7,
        sigma_s=1e-6,
        sigma_p=1e-6,
        cor_x=0,
        cor_y=0,
        cor_s=0,
        energy=1e8,
    ):
        return cls(
            mu=torch.tensor([mu_x, mu_xp, mu_y, mu_yp, 0, 0, 1], dtype=torch.float32),
            cov=torch.tensor(
                [
                    [sigma_x**2, cor_x, 0, 0, 0, 0, 0],
                    [cor_x, sigma_xp**2, 0, 0, 0, 0, 0],
                    [0, 0, sigma_y**2, cor_y, 0, 0, 0],
                    [0, 0, cor_y, sigma_yp**2, 0, 0, 0],
                    [0, 0, 0, 0, sigma_s**2, cor_s, 0],
                    [0, 0, 0, 0, cor_s, sigma_p**2, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                ],
                dtype=torch.float32,
            ),
            energy=energy,
        )

    @classmethod
    def from_ocelot(cls, parray):
        mu = torch.ones(7)
        mu[:6] = torch.tensor(parray.rparticles.mean(axis=1), dtype=torch.float32)

        cov = torch.zeros(7, 7)
        cov[:6, :6] = torch.tensor(np.cov(parray.rparticles), dtype=torch.float32)

        energy = 1e9 * parray.E

        return cls(mu=mu, cov=cov, energy=energy)

    @classmethod
    def from_astra(cls, path, **kwargs):
        """Load an Astra particle distribution as a Cheetah Beam."""
        particles, energy = from_astrabeam(path)
        mu = torch.ones(7)
        mu[:6] = torch.tensor(particles.mean(axis=0), dtype=torch.float32)

        cov = torch.zeros(7, 7)
        cov[:6, :6] = torch.tensor(np.cov(particles.transpose()), dtype=torch.float32)

        return cls(mu=mu, cov=cov, energy=energy)

    def transformed_to(
        self,
        mu_x=None,
        mu_xp=None,
        mu_y=None,
        mu_yp=None,
        sigma_x=None,
        sigma_xp=None,
        sigma_y=None,
        sigma_yp=None,
        sigma_s=None,
        sigma_p=None,
        energy=None,
    ):
        """
        Create version of this beam that is transformed to new beam parameters.

        Parameters
        ----------
        n : int, optional
            Number of particles to generate.
        mu_x : float, optional
            Center of the particle distribution on x in meters.
        mu_xp : float, optional
            Center of the particle distribution on px in meters.
        mu_y : float, optional
            Center of the particle distribution on y in meters.
        mu_yp : float, optional
            Center of the particle distribution on py in meters.
        sigma_x : float, optional
            Sgima of the particle distribution in x direction in meters.
        sigma_xp : float, optional
            Sgima of the particle distribution in px direction in meters.
        sigma_y : float, optional
            Sgima of the particle distribution in y direction in meters.
        sigma_yp : float, optional
            Sgima of the particle distribution in py direction in meters.
        sigma_s : float, optional
            Sgima of the particle distribution in s direction in meters.
        sigma_p : float, optional
            Sgima of the particle distribution in p direction in meters.
        energy : float, optional
            Energy of the beam in eV.
        """
        mu_x = mu_x if mu_x != None else self.mu_x
        mu_xp = mu_xp if mu_xp != None else self.mu_xp
        mu_y = mu_y if mu_y != None else self.mu_y
        mu_yp = mu_yp if mu_yp != None else self.mu_yp
        sigma_x = sigma_x if sigma_x != None else self.sigma_x
        sigma_xp = sigma_xp if sigma_xp != None else self.sigma_xp
        sigma_y = sigma_y if sigma_y != None else self.sigma_y
        sigma_yp = sigma_yp if sigma_yp != None else self.sigma_yp
        sigma_s = sigma_s if sigma_s != None else self.sigma_s
        sigma_p = sigma_p if sigma_p != None else self.sigma_p
        energy = energy if energy != None else self.energy

        return self.__class__.from_parameters(
            mu_x=mu_x,
            mu_xp=mu_xp,
            mu_y=mu_y,
            mu_yp=mu_yp,
            sigma_x=sigma_x,
            sigma_xp=sigma_xp,
            sigma_y=sigma_y,
            sigma_yp=sigma_yp,
            sigma_s=sigma_s,
            sigma_p=sigma_p,
            energy=energy,
        )

    @property
    def mu_x(self):
        return self._mu[0]

    @property
    def sigma_x(self):
        return torch.sqrt(self._cov[0, 0])

    @property
    def mu_xp(self):
        return self._mu[1]

    @property
    def sigma_xp(self):
        return torch.sqrt(self._cov[1, 1])

    @property
    def mu_y(self):
        return self._mu[2]

    @property
    def sigma_y(self):
        return torch.sqrt(self._cov[2, 2])

    @property
    def mu_yp(self):
        return self._mu[3]

    @property
    def sigma_yp(self):
        return torch.sqrt(self._cov[3, 3])

    @property
    def mu_s(self):
        return self._mu[4]

    @property
    def sigma_s(self):
        return torch.sqrt(self._cov[4, 4])

    @property
    def mu_p(self):
        return self._mu[5]

    @property
    def sigma_p(self):
        return torch.sqrt(self._cov[5, 5])

    def __repr__(self):
        return f"{self.__class__.__name__}(mu_x={self.mu_x:.6f}, mu_xp={self.mu_xp:.6f}, mu_y={self.mu_y:.6f}, mu_yp={self.mu_yp:.6f}, sigma_x={self.sigma_x:.6f}, sigma_xp={self.sigma_xp:.6f}, sigma_y={self.sigma_y:.6f}, sigma_yp={self.sigma_yp:.6f}, sigma_s={self.sigma_s:.6f}, sigma_p={self.sigma_p:.6f}, energy={self.energy:.3f})"


class ParticleBeam:
    """
    Beam of charged particles, where each particle is simulated.

    Parameters
    ----------
    particles : torch.Tensor
        List of 7-dimensional particle vectors.
    energy : float
        Energy of the beam in eV.
    device : string
        Device to move the beam's particle array to. If set to `"auto"` a CUDA GPU is
        selected if available. The CPU is used otherwise.
    """

    def __init__(self, particles, energy, device="auto"):
        assert (
            len(particles) > 0 and particles.shape[1] == 7
        ), "Particle vectors must be 7-dimensional."

        if device == "auto":
            device = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = device

        self.particles = (
            particles.to(device) if isinstance(particles, torch.Tensor) else particles
        )

        self.energy = energy

    @classmethod
    def from_parameters(
        cls,
        n=100000,
        mu_x=0,
        mu_y=0,
        mu_xp=0,
        mu_yp=0,
        sigma_x=175e-9,
        sigma_y=175e-9,
        sigma_xp=2e-7,
        sigma_yp=2e-7,
        sigma_s=1e-6,
        sigma_p=1e-6,
        cor_x=0,
        cor_y=0,
        cor_s=0,
        energy=1e8,
        device="auto",
    ):
        """
        Generate Cheetah Beam of random particles.

        Parameters
        ----------
        n : int, optional
            Number of particles to generate.
        mu_x : float, optional
            Center of the particle distribution on x in meters.
        mu_y : float, optional
            Center of the particle distribution on y in meters.
        mu_xp : float, optional
            Center of the particle distribution on px in meters.
        mu_yp : float, optional
            Center of the particle distribution on py in meters.
        sigma_x : float, optional
            Sgima of the particle distribution in x direction in meters.
        sigma_y : float, optional
            Sgima of the particle distribution in y direction in meters.
        sigma_xp : float, optional
            Sgima of the particle distribution in px direction in meters.
        sigma_yp : float, optional
            Sgima of the particle distribution in py direction in meters.
        sigma_s : float, optional
            Sgima of the particle distribution in s direction in meters.
        sigma_p : float, optional
            Sgima of the particle distribution in p direction in meters.
        cor_x : float, optional
            Correlation between x and xp.
        cor_y : float, optional
            Correlation between y and yp.
        cor_s : float, optional
            Correlation between s and p.
        energy : float, optional
            Energy of the beam in eV.
        device : string
            Device to move the beam's particle array to. If set to `"auto"` a CUDA GPU
            is selected if available. The CPU is used otherwise.
        """
        mean = torch.tensor([mu_x, mu_xp, mu_y, mu_yp, 0, 0], dtype=torch.float32)
        cov = torch.tensor(
            [
                [sigma_x**2, cor_x, 0, 0, 0, 0],
                [cor_x, sigma_xp**2, 0, 0, 0, 0],
                [0, 0, sigma_y**2, cor_y, 0, 0],
                [0, 0, cor_y, sigma_yp**2, 0, 0],
                [0, 0, 0, 0, sigma_s**2, cor_s],
                [0, 0, 0, 0, cor_s, sigma_p**2],
            ],
            dtype=torch.float32,
        )

        particles = torch.ones((n, 7), dtype=torch.float32)
        distribution = MultivariateNormal(mean, covariance_matrix=cov)
        particles[:, :6] = distribution.sample((n,))

        return cls(particles, energy, device=device)

    @classmethod
    def make_linspaced(
        cls,
        n=10,
        mu_x=0,
        mu_y=0,
        mu_xp=0,
        mu_yp=0,
        sigma_x=175e-9,
        sigma_y=175e-9,
        sigma_xp=2e-7,
        sigma_yp=2e-7,
        sigma_s=0,
        sigma_p=0,
        energy=1e8,
        device="auto",
    ):
        """
        Generate Cheetah Beam of *n* linspaced particles.

        Parameters
        ----------
        n : int, optional
            Number of particles to generate.
        mu_x : float, optional
            Center of the particle distribution on x in meters.
        mu_y : float, optional
            Center of the particle distribution on y in meters.
        mu_px : float, optional
            Center of the particle distribution on px in meters.
        mu_py : float, optional
            Center of the particle distribution on py in meters.
        sigma_x : float, optional
            Sgima of the particle distribution in x direction in meters.
        sigma_y : float, optional
            Sgima of the particle distribution in y direction in meters.
        sigma_xp : float, optional
            Sgima of the particle distribution in px direction in meters.
        sigma_yp : float, optional
            Sgima of the particle distribution in py direction in meters.
        sigma_s : float, optional
            Sgima of the particle distribution in s direction in meters.
        sigma_p : float, optional
            Sgima of the particle distribution in p direction in meters.
        energy : float, optional
            Energy of the beam in eV.
        device : string
            Device to move the beam's particle array to. If set to `"auto"` a CUDA GPU
            is selected if available. The CPU is used otherwise.
        """
        particles = torch.ones((n, 7), dtype=torch.float32)

        particles[:, 0] = torch.linspace(
            mu_x - sigma_x, mu_x + sigma_x, n, dtype=torch.float32
        )
        particles[:, 1] = torch.linspace(
            mu_xp - sigma_xp, mu_xp + sigma_xp, n, dtype=torch.float32
        )
        particles[:, 2] = torch.linspace(
            mu_y - sigma_y, mu_y + sigma_y, n, dtype=torch.float32
        )
        particles[:, 3] = torch.linspace(
            mu_yp - sigma_yp, mu_yp + sigma_yp, n, dtype=torch.float32
        )
        particles[:, 4] = torch.linspace(-sigma_s, sigma_s, n, dtype=torch.float32)
        particles[:, 5] = torch.linspace(-sigma_p, sigma_p, n, dtype=torch.float32)

        return cls(particles, energy, device=device)

    @classmethod
    def from_ocelot(cls, parray, device="auto"):
        """
        Convert an Ocelot ParticleArray `parray` to a Cheetah Beam.
        """
        n = parray.rparticles.shape[1]
        particles = torch.ones((n, 7))
        particles[:, :6] = torch.tensor(
            parray.rparticles.transpose(), dtype=torch.float32
        )

        return cls(particles, 1e9 * parray.E, device=device)

    @classmethod
    def from_astra(cls, path, **kwargs):
        """Load an Astra particle distribution as a Cheetah Beam."""
        particles, energy = from_astrabeam(path)
        particles_7d = torch.ones((particles.shape[0], 7))
        particles_7d[:, :6] = torch.from_numpy(particles)
        return cls(particles_7d, energy, **kwargs)

    def transformed_to(
        self,
        mu_x=None,
        mu_y=None,
        mu_xp=None,
        mu_yp=None,
        sigma_x=None,
        sigma_y=None,
        sigma_xp=None,
        sigma_yp=None,
        sigma_s=None,
        sigma_p=None,
        energy=None,
    ):
        """
        Create version of this beam that is transformed to new beam parameters.

        Parameters
        ----------
        n : int, optional
            Number of particles to generate.
        mu_x : float, optional
            Center of the particle distribution on x in meters.
        mu_y : float, optional
            Center of the particle distribution on y in meters.
        mu_xp : float, optional
            Center of the particle distribution on px in meters.
        mu_yp : float, optional
            Center of the particle distribution on py in meters.
        sigma_x : float, optional
            Sgima of the particle distribution in x direction in meters.
        sigma_y : float, optional
            Sgima of the particle distribution in y direction in meters.
        sigma_xp : float, optional
            Sgima of the particle distribution in px direction in meters.
        sigma_yp : float, optional
            Sgima of the particle distribution in py direction in meters.
        sigma_s : float, optional
            Sgima of the particle distribution in s direction in meters.
        sigma_p : float, optional
            Sgima of the particle distribution in p direction in meters.
        energy : float, optional
            Energy of the beam in eV.
        device : string
            Device to move the beam's particle array to. If set to `"auto"` a CUDA GPU
            is selected if available. The CPU is used otherwise.
        """
        mu_x = mu_x if mu_x != None else self.mu_x
        mu_y = mu_y if mu_y != None else self.mu_y
        mu_xp = mu_xp if mu_xp != None else self.mu_xp
        mu_yp = mu_yp if mu_yp != None else self.mu_yp
        sigma_x = sigma_x if sigma_x != None else self.sigma_x
        sigma_y = sigma_y if sigma_y != None else self.sigma_y
        sigma_xp = sigma_xp if sigma_xp != None else self.sigma_xp
        sigma_yp = sigma_yp if sigma_yp != None else self.sigma_yp
        sigma_s = sigma_s if sigma_s != None else self.sigma_s
        sigma_p = sigma_p if sigma_p != None else self.sigma_p
        energy = energy if energy != None else self.energy

        new_mu = torch.tensor(
            [mu_x, mu_xp, mu_y, mu_yp, 0, 0], dtype=torch.float32, device=self.device
        )
        new_sigma = torch.tensor(
            [sigma_x, sigma_xp, sigma_y, sigma_yp, sigma_s, sigma_p],
            dtype=torch.float32,
            device=self.device,
        )

        old_mu = torch.tensor(
            [self.mu_x, self.mu_xp, self.mu_y, self.mu_yp, 0, 0],
            dtype=torch.float32,
            device=self.device,
        )
        old_sigma = torch.tensor(
            [
                self.sigma_x,
                self.sigma_xp,
                self.sigma_y,
                self.sigma_yp,
                self.sigma_s,
                self.sigma_p,
            ],
            dtype=torch.float32,
            device=self.device,
        )

        phase_space = self.particles[:, :6]
        phase_space = (phase_space - old_mu) / old_sigma * new_sigma + new_mu

        particles = torch.ones_like(
            self.particles, dtype=torch.float32, device=self.device
        )
        particles[:, :6] = phase_space

        return self.__class__(particles=particles, energy=energy)

    def __len__(self):
        return self.n

    @property
    def n(self):
        return len(self.particles)

    @property
    def xs(self):
        return self.particles[:, 0] if self is not Beam.empty else None

    @xs.setter
    def xs(self, value):
        self.particles[:, 0] = value

    @property
    def mu_x(self):
        return float(self.xs.mean()) if self is not Beam.empty else None

    @property
    def sigma_x(self):
        return float(self.xs.std()) if self is not Beam.empty else None

    @property
    def xps(self):
        return self.particles[:, 1] if self is not Beam.empty else None

    @xps.setter
    def xps(self, value):
        self.particles[:, 1] = value

    @property
    def mu_xp(self):
        return float(self.xps.mean()) if self is not Beam.empty else None

    @property
    def sigma_xp(self):
        return float(self.xps.std()) if self is not Beam.empty else None

    @property
    def ys(self):
        return self.particles[:, 2] if self is not Beam.empty else None

    @ys.setter
    def ys(self, value):
        self.particles[:, 2] = value

    @property
    def mu_y(self):
        return float(self.ys.mean()) if self is not Beam.empty else None

    @property
    def sigma_y(self):
        return float(self.ys.std()) if self is not Beam.empty else None

    @property
    def yps(self):
        return self.particles[:, 3] if self is not Beam.empty else None

    @yps.setter
    def yps(self, value):
        self.particles[:, 3] = value

    @property
    def mu_yp(self):
        return float(self.yps.mean()) if self is not Beam.empty else None

    @property
    def sigma_yp(self):
        return float(self.yps.std()) if self is not Beam.empty else None

    @property
    def ss(self):
        return self.particles[:, 4] if self is not Beam.empty else None

    @ss.setter
    def ss(self, value):
        self.particles[:, 4] = value

    @property
    def mu_s(self):
        return float(self.ss.mean()) if self is not Beam.empty else None

    @property
    def sigma_s(self):
        return float(self.ss.std()) if self is not Beam.empty else None

    @property
    def ps(self):
        return self.particles[:, 5] if self is not Beam.empty else None

    @ps.setter
    def ps(self, value):
        self.particles[:, 5] = value

    @property
    def mu_p(self):
        return float(self.ps.mean()) if self is not Beam.empty else None

    @property
    def sigma_p(self):
        return float(self.ps.std()) if self is not Beam.empty else None

    def __repr__(self):
        return f"{self.__class__.__name__}(n={self.n}, mu_x={self.mu_x:.6f}, mu_xp={self.mu_xp:.6f}, mu_y={self.mu_y:.6f}, mu_yp={self.mu_yp:.6f}, sigma_x={self.sigma_x:.6f}, sigma_xp={self.sigma_xp:.6f}, sigma_y={self.sigma_y:.6f}, sigma_yp={self.sigma_yp:.6f}, sigma_s={self.sigma_s:.6f}, sigma_p={self.sigma_p:.6f}, energy={self.energy:.3f})"



from pathlib import Path

from setuptools import setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="cheetah-accelerator",
    version="0.5.17",
    author="Jan Kaiser & Oliver Stein",
    author_email="jan.kaiser@desy.de",
    url="https://github.com/desy-ml/cheetah",
    description=(
        "Fast particle accelerator optics simulation for reinforcement learning and"
        " optimisation applications."
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=["cheetah"],
    install_requires=[
        "torch",
        "matplotlib",
        "numpy",
        "scipy",
    ],
)



import numpy as np
import torch

from cheetah import accelerator as acc

# electron mass in eV
m_e_eV = 510998.8671


def from_astrabeam(path: str):
    """
    Read from a ASTRA beam distribution, and prepare for conversion to a Cheetah
    ParticleBeam or ParameterBeam.

    Adapted from the implementation in ocelot:
    https://github.com/ocelot-collab/ocelot/blob/master/ocelot/adaptors/astra2ocelot.py

    Parameters
    ----------
    path : str
        Path to the ASTRA beam distribution file.

    Returns
    -------
    particles : np.ndarray
        Particle 6D phase space information
    energy : float
        Mean energy of the particle beam

    """
    P0 = np.loadtxt(path)

    # remove lost particles
    inds = np.argwhere(P0[:, 9] > 0)
    inds = inds.reshape(inds.shape[0])

    P0 = P0[inds, :]
    n_particles = P0.shape[0]

    # s_ref = P0[0, 2]
    Pref = P0[0, 5]

    xp = P0[:, :6]
    xp[0, 2] = 0.0
    xp[0, 5] = 0.0

    gamref = np.sqrt((Pref / m_e_eV) ** 2 + 1)
    # energy in eV: E = gamma * m_e
    energy = gamref * m_e_eV

    n_particles = xp.shape[0]
    particles = np.zeros((n_particles, 6))

    u = np.c_[xp[:, 3], xp[:, 4], xp[:, 5] + Pref]
    gamma = np.sqrt(1 + np.sum(u * u, 1) / m_e_eV**2)
    beta = np.sqrt(1 - gamma**-2)
    betaref = np.sqrt(1 - gamref**-2)

    p0 = np.linalg.norm(u, 2, 1).reshape((n_particles, 1))

    u = u / p0
    cdt = -xp[:, 2] / (beta * u[:, 2])
    particles[:, 0] = xp[:, 0] + beta * u[:, 0] * cdt
    particles[:, 2] = xp[:, 1] + beta * u[:, 1] * cdt
    particles[:, 4] = cdt
    particles[:, 1] = xp[:, 3] / Pref
    particles[:, 3] = xp[:, 4] / Pref
    particles[:, 5] = (gamma / gamref - 1) / betaref

    return particles, energy


def ocelot2cheetah(element, warnings=True):
    """
    Translate an Ocelot element to a Cheetah element.

    Parameters
    ----------
    element : ocelot.Element
        Ocelot element object representing an element of particle accelerator.

    Returns
    -------
    cheetah.Element
        Cheetah element object representing an element of particle accelerator.

    Notes
    -----
    Object not supported by Cheetah are translated to drift sections. Screen objects are
    created only from `ocelot.Monitor` objects when the string "SCR" in their `id`
    attribute. Their screen properties are always set to default values and most likely
    need adjusting afterwards. BPM objects are only created from `ocelot.Monitor`
    objects when their id has a substring "BPM".
    """
    try:
        import ocelot as oc
    except ImportError:
        raise ImportError(
            """To use the ocelot2cheetah lattice converter, Ocelot must be first 
        installed, see https://github.com/ocelot-collab/ocelot """
        )

    if isinstance(element, oc.Drift):
        return acc.Drift(element.l, name=element.id)
    elif isinstance(element, oc.Quadrupole):
        return acc.Quadrupole(element.l, element.k1, name=element.id)
    elif isinstance(element, oc.Hcor):
        return acc.HorizontalCorrector(element.l, element.angle, name=element.id)
    elif isinstance(element, oc.Vcor):
        return acc.VerticalCorrector(element.l, element.angle, name=element.id)
    elif isinstance(element, oc.Cavity):
        return acc.Cavity(element.l, name=element.id)
    elif isinstance(element, oc.Monitor) and "BSC" in element.id:
        if warnings:
            print(
                "WARNING: Diagnostic screen was converted with default screen"
                " properties."
            )
        return acc.Screen((2448, 2040), (3.5488e-6, 2.5003e-6), name=element.id)
    elif isinstance(element, oc.Monitor) and "BPM" in element.id:
        return acc.BPM(name=element.id)
    elif isinstance(element, oc.Undulator):
        return acc.Undulator(element.l, name=element.id)
    else:
        if warnings:
            print(
                f"WARNING: Unknown element {element.id}, replacing with drift section."
            )
        return acc.Drift(element.l, name=element.id)


def subcell_of_ocelot(cell, start, end):
    """Extract a subcell `[start, end]` from an Ocelot cell."""
    subcell = []
    is_in_subcell = False
    for el in cell:
        if el.id == start:
            is_in_subcell = True
        if is_in_subcell:
            subcell.append(el)
        if el.id == end:
            break

    return subcell
